
#define _WIN32_WINNT 0x0400
#define _WIN32_IE 0x0500

#include <windows.h>
#include <afxres.h>
#include <commctrl.h>
extern "C" {
#include <powrprof.h>
}
#include <shlwapi.h>
#include <stdarg.h>
#include <tchar.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iterator>
#include <limits>
#include <list>
#include <queue>
#include <sstream>
#include <string>
#include <vector>
#include <stdext/CLI.h>
#include <stdext/utils.h>
#include "resource.h"

using namespace std;
using namespace stdext;

#undef max
#define APP_VERSION "2.2 beta"

namespace std {
typedef basic_string<TCHAR> _tstring;
}

static FILETIME filetime_add_seconds(FILETIME ft, int seconds) {
	__asm {
		mov eax, 10000000
		imul seconds
		add eax, DWORD PTR [ft]
		adc edx, DWORD PTR [ft + 4]
	}
}

static int filetime_subtract_get_seconds(FILETIME ftMinuend, FILETIME ftSubtrahend) {
	__asm {
		mov eax, DWORD PTR [ftMinuend]
		mov edx, DWORD PTR [ftMinuend + 4]
		mov ecx, 10000000
		sub eax, DWORD PTR [ftSubtrahend]
		sbb edx, DWORD PTR [ftSubtrahend + 4]
		idiv ecx
	}
}

static void format_seconds(int seconds, string &str) {
	stringstream sstr;
	if (seconds < 0) {
		seconds = -seconds;
		sstr << '-';
	}
	sstr.fill('0');
	sstr << (seconds / 3600) << ':';
	sstr.width(2);
	sstr << ((seconds / 60) % 60) << ':';
	sstr.width(2);
	sstr << (seconds % 60);
	str.assign(sstr.str());
}

static void get_short_date(const SYSTEMTIME *pst, string &str) {
	int len = GetDateFormatA(LOCALE_SYSTEM_DEFAULT, LOCALE_NOUSEROVERRIDE | LOCALE_USE_CP_ACP, pst, NULL, NULL, 0);
	str.reserve(len);
	str.resize(len - 1);
	GetDateFormatA(LOCALE_SYSTEM_DEFAULT, LOCALE_NOUSEROVERRIDE | LOCALE_USE_CP_ACP, pst, NULL, const_cast<char *>(str.data()), len);
}

static void get_short_time(const SYSTEMTIME *pst, string &str) {
	int len = GetTimeFormatA(LOCALE_SYSTEM_DEFAULT, LOCALE_NOUSEROVERRIDE | LOCALE_USE_CP_ACP, pst, NULL, NULL, 0);
	str.reserve(len);
	str.resize(len - 1);
	GetTimeFormatA(LOCALE_SYSTEM_DEFAULT, LOCALE_NOUSEROVERRIDE | LOCALE_USE_CP_ACP, pst, NULL, const_cast<char *>(str.data()), len);
}

static void xml_special_chars(const string &in, string &out) {
	stringbuf ibuf(in), obuf;
	int c;
	while (!stringbuf::traits_type::eq_int_type(c = ibuf.sbumpc(), stringbuf::traits_type::eof())) {
		if (c == '&') {
			obuf.sputn("&amp;", 5);
		}
		else if (c == '<') {
			obuf.sputn("&lt;", 4);
		}
		else if (c == '>') {
			obuf.sputn("&gt;", 4);
		}
		else if (c == '"') {
			obuf.sputn("&quot;", 6);
		}
		else if (c == '\'') {
			obuf.sputn("&apos;", 6);
		}
		else {
			obuf.sputc(stringbuf::traits_type::to_char_type(c));
		}
	}
	out.assign(obuf.str());
}

class SageWake {
	friend int _tmain(int argc, TCHAR **argv);

	class Range {
	public:
		BYTE low;
		BYTE high;

		Range() { }

		Range(BYTE low, BYTE high) : low(low), high(high) { }

		bool parse(LPCTSTR psz) {
			size_t i;
			for (i = 0; (psz[i] >= '0') && (psz[i] <= '9'); i++);
			if ((psz[i] != '-') || !((psz[i + 1] >= '0') && (psz[i + 1] <= '9'))) {
				return false;
			}
			low = static_cast<BYTE>(_tstoi(psz));
			high = static_cast<BYTE>(_tstoi(psz + i + 1));
			return true;
		}

		bool in_range(BYTE b) const {
			return (low <= high) ? ((b >= low) && (b < high)) : ((b >= low) || (b < high));
		}

		static bool in_any_range(BYTE b, const list<Range> &ranges) {
			return find_if(ranges.begin(), ranges.end(), bind2nd(mem_fun_ref(&Range::in_range), b)) != ranges.end();
		}
	};

	class Options {
	private:
		static bool parse_ranges(LPCTSTR pszRanges, list<Range> &ranges) {
			for (;;) {
				Range range;
				if (!range.parse(pszRanges)) {
					return false;
				}
				ranges.push_back(range);
				pszRanges = _tcschr(pszRanges, ',');
				if (pszRanges == NULL) {
					return true;
				}
				++pszRanges;
			}
		}

	public:
		bool bAppendLog;
		bool bNoResuspend;
		bool bUseHiber;
		int iLeadTime;
		int iWarnTime;
		list<_tstring> lstExecute;
		list<_tstring> lstTerminate;
		LPCTSTR pszLogFile;
		LPCTSTR pszSchedFile;
		list<Range> lstNoFavorite;
		list<Range> lstNoIR;
		list<Range> lstNoManual;

		Options() : bAppendLog(false), bNoResuspend(false), bUseHiber(false), iLeadTime(60), iWarnTime(60), pszLogFile(NULL), pszSchedFile(TEXT("schedule.txt")) {
		}

		bool initialize(int argc, const TCHAR **argv) {
			const TCHAR *pszNoManual = TEXT(""), *pszNoIR = TEXT(""), *pszNoFavorite = TEXT(""), *pszTerminate = NULL, *pszExecute = NULL;
			const CLI::Option opts[] = {
				{ CLI::STRING, 'M', TEXT("nomanual"), &pszNoManual },
				{ CLI::STRING, 'I', TEXT("noir"), &pszNoIR },
				{ CLI::STRING, 'F', TEXT("nofavorite"), &pszNoFavorite },
				{ CLI::INTEGER, 't', TEXT("leadtime"), &iLeadTime },
				{ CLI::STRING, 's', TEXT("schedfile"), &pszSchedFile },
				{ CLI::STRING, 'l', TEXT("logfile"), &pszLogFile },
				{ CLI::BOOLEAN, 'a', TEXT("appendlog"), &bAppendLog },
				{ CLI::STRING, 'T', TEXT("terminate"), &pszTerminate },
				{ CLI::STRING, 'X', TEXT("execute"), &pszExecute },
				{ CLI::BOOLEAN, 'R', TEXT("noresuspend"), &bNoResuspend },
				{ CLI::INTEGER, 'w', TEXT("warntime"), &iWarnTime },
				{ CLI::BOOLEAN, 'h', TEXT("usehiber"), &bUseHiber }
			};
			const TCHAR *err;
			if (CLI::parse(opts, countof(opts), argc, argv, &err) == 0) {
				TCHAR buf[256];
				CLI::format_error(err, buf, sizeof(buf));
				MessageBox(NULL, buf, TEXT("SageWake"), MB_ICONEXCLAMATION);
				return false;
			}
			Range range24hr(0, 24);
			if (pszNoManual != NULL) {
				if ((*pszNoManual != '\0') && !parse_ranges(pszNoManual, lstNoManual)) {
					MessageBox(NULL, TEXT("There was a problem with your --nomanual ranges."), TEXT("SageWake"), MB_ICONEXCLAMATION);
					return false;
				}
			}
			else {
				lstNoManual.push_back(range24hr);
			}
			if (pszNoIR != NULL) {
				if ((*pszNoIR != '\0') && !parse_ranges(pszNoIR, lstNoIR)) {
					MessageBox(NULL, TEXT("There was a problem with your --noir ranges."), TEXT("SageWake"), MB_ICONEXCLAMATION);
					return false;
				}
			}
			else {
				lstNoIR.push_back(range24hr);
			}
			if (pszNoFavorite != NULL) {
				if ((*pszNoFavorite != '\0') && !parse_ranges(pszNoFavorite, lstNoFavorite)) {
					MessageBox(NULL, TEXT("There was a problem with your --nofavorite ranges."), TEXT("SageWake"), MB_ICONEXCLAMATION);
					return false;
				}
			}
			else {
				lstNoFavorite.push_back(range24hr);
			}
			if (pszTerminate != NULL) {
				explode(TEXT(","), pszTerminate, lstTerminate);
			}
			if (pszExecute != NULL) {
				explode(TEXT(","), pszExecute, lstExecute);
			}
			return true;
		}
	} opts;

	class Log {
	private:
		HANDLE hLogFile;

	public:
		Log() : hLogFile(INVALID_HANDLE_VALUE) {
		}

		~Log() {
			close();
		}

		BOOL open(LPCTSTR pszLogFile, bool bAppend) {
			close();
			if (bAppend && (GetFileAttributes(pszLogFile) == INVALID_FILE_ATTRIBUTES)) {
				bAppend = false;
			}
			hLogFile = CreateFile(pszLogFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, bAppend ? OPEN_ALWAYS : CREATE_ALWAYS, NULL, NULL);
			if (hLogFile != INVALID_HANDLE_VALUE) {
				if (!bAppend) {
					DWORD dw;
					LPCSTR pszInit = "<?xml version=\"1.0\" encoding=\"Windows-1252\" standalone=\"yes\" ?>\r\n<SageWakeLog version=\"" APP_VERSION "\">\r\n</SageWakeLog>";
					WriteFile(hLogFile, pszInit, (DWORD)strlen(pszInit), &dw, NULL);
				}
				return event("<SageWakeStart />");
			}
			else {
				return FALSE;
			}
		}

		bool close() {
			if (hLogFile != INVALID_HANDLE_VALUE) {
				event("<SageWakeStop />");
				CloseHandle(hLogFile), hLogFile = INVALID_HANDLE_VALUE;
				return true;
			}
			else {
				return false;
			}
		}

		bool is_open() const {
			return (hLogFile != INVALID_HANDLE_VALUE);
		}

		BOOL event(LPCSTR psz) const {
			DWORD dw;
			string str;
			if (hLogFile != INVALID_HANDLE_VALUE) {
				SetFilePointer(hLogFile, -14, NULL, FILE_END);
				stringstream stm;
				string buf;
				get_short_date(NULL, buf);
				stm << "<Event time=\"" << buf;
				get_short_time(NULL, buf);
				stm << ' ' << buf << "\">" << psz << "</Event>\r\n</SageWakeLog>";
				return WriteFile(hLogFile, stm.str().data(), (DWORD)stm.tellp(), &dw, NULL);
			}
			else {
				return FALSE;
			}
		}
	} log;

	class Show {
	private:
		static bool sagetime_to_filetime(LPCSTR pszSageTime, FILETIME *pft)
		{
			SYSTEMTIME stLocal;
			GetLocalTime(&stLocal);
			LPCSTR psz = strchr(pszSageTime, '/');
			if (psz == NULL) {
				return false;
			}
			stLocal.wDay = static_cast<WORD>(atoi(psz + 1));
			while (*(--psz) != ' ');
			if (psz == NULL) {
				return false;
			}
			WORD w = static_cast<WORD>(atoi(psz + 1));
			if ((w == 1) && (stLocal.wMonth == 12)) {
				stLocal.wYear++;
			}
			else if ((w == 12) && (stLocal.wMonth == 1)) {
				stLocal.wYear--;
			}
			stLocal.wMonth = w;
			psz = strchr(psz, ':');
			if (psz == NULL) {
				return false;
			}
			stLocal.wMinute = static_cast<WORD>(atoi(psz + 1));
			while (*(--psz) != ' ');
			if (psz == NULL) {
				return false;
			}
			stLocal.wHour = static_cast<WORD>(atoi(psz + 1));
			stLocal.wSecond = 0;
			stLocal.wMilliseconds = 0;
			FILETIME ftLocal;
			return SystemTimeToFileTime(&stLocal, &ftLocal) && LocalFileTimeToFileTime(&ftLocal, pft);
		}

	public:
		string strEncoder;
		FILETIME ftStart;
		FILETIME ftStop;
		string strChannelName;
		USHORT iChannelNum;
		string strTitle;
		bool bManual;
		bool bFavorite;

		bool parse(const char *psz) {
			list<string> lstFields;
			explode("\t", psz, lstFields);
			if (lstFields.size() < 9) {
				return false;
			}
			list<string>::const_iterator it = lstFields.begin();
			strEncoder.assign(*it++);
			if (!sagetime_to_filetime((it++)->c_str(), &ftStart) || !sagetime_to_filetime((it++)->c_str(), &ftStop)) {
				return false;
			}
			++it;
			strChannelName.assign(*it++);
			iChannelNum = static_cast<USHORT>(atoi((it++)->c_str()));
			strTitle.assign(*it++);
			bManual = (*it++ == "true");
			bFavorite = (*it++ == "true");
			return true;
		}

		bool is_selected(const Options &opts) const {
			FILETIME ftLocalStart;
			FileTimeToLocalFileTime(&ftStart, &ftLocalStart);
			SYSTEMTIME stLocalStart;
			FileTimeToSystemTime(&ftLocalStart, &stLocalStart);
			return ((!bManual || !Range::in_any_range((BYTE)stLocalStart.wHour, opts.lstNoManual)) && (!bFavorite || !Range::in_any_range((BYTE)stLocalStart.wHour, opts.lstNoFavorite)) && (bManual || bFavorite || !Range::in_any_range((BYTE)stLocalStart.wHour, opts.lstNoIR)));
		}

		void to_gui_string(string &str, const Options &opts, bool for_schedule_list) const {
			FILETIME ftStartLocal;
			FileTimeToLocalFileTime(&ftStart, &ftStartLocal);
			SYSTEMTIME stStartLocal;
			FileTimeToSystemTime(&ftStartLocal, &stStartLocal);
			int len = GetTimeFormat(LOCALE_USER_DEFAULT, TIME_NOSECONDS, &stStartLocal, NULL, NULL, 0);
			if (for_schedule_list) {
				str = is_selected(opts) ? "*\t" : " \t";
				str.reserve(len + 2);
				str.resize(len + 1);
				GetTimeFormatA(LOCALE_USER_DEFAULT, TIME_NOSECONDS, &stStartLocal, NULL, const_cast<char *>(str.data() + 2), len);
				str.push_back('\t');
				str.append(strTitle);
			}
			else {
				str.reserve(len);
				str.resize(len - 1);
				GetTimeFormatA(LOCALE_USER_DEFAULT, TIME_NOSECONDS, &stStartLocal, NULL, const_cast<char *>(str.data()), len);
				str.push_back(' ');
				str.append(strTitle);
			}
		}

		void to_xml(string &str) const {
			FILETIME ftLocal;
			SYSTEMTIME stLocalStart, stLocalStop;
			FileTimeToLocalFileTime(&ftStart, &ftLocal);
			FileTimeToSystemTime(&ftLocal, &stLocalStart);
			FileTimeToLocalFileTime(&ftStop, &ftLocal);
			FileTimeToSystemTime(&ftLocal, &stLocalStop);
			stringstream stm;
			string buf;
			xml_special_chars(strTitle, buf);
			stm << "<Show title=\"" << buf;
			get_short_date(&stLocalStart, buf);
			stm << "\" start=\"" << buf;
			get_short_time(&stLocalStart, buf);
			stm << ' ' << buf;
			get_short_date(&stLocalStop, buf);
			stm << "\" stop=\"" << buf;
			get_short_time(&stLocalStop, buf);
			stm << ' ' << buf << "\" channel=\"" << iChannelNum << "\" />";
			str = stm.str();
		}

		bool operator == (const Show &rhs) const {
			return (*(UINT64 *)&ftStart == *(UINT64 *)&rhs.ftStart) && (*(UINT64 *)&ftStop == *(UINT64 *)&rhs.ftStop) && (iChannelNum == rhs.iChannelNum) && (strTitle == rhs.strTitle);
		}
		
		bool operator < (const Show &rhs) const {
			return ((*(UINT64 *)&ftStart < *(UINT64 *)&rhs.ftStart) || ((*(UINT64 *)&ftStart == *(UINT64 *)&rhs.ftStart) && ((*(UINT64 *)&ftStop < *(UINT64 *)&rhs.ftStop) || ((*(UINT64 *)&ftStop == *(UINT64 *)&rhs.ftStop) && ((iChannelNum < rhs.iChannelNum) || ((iChannelNum == rhs.iChannelNum) && (strTitle < rhs.strTitle)))))));
		}
		
		bool operator > (const Show &rhs) const {
			return ((*(UINT64 *)&ftStart > *(UINT64 *)&rhs.ftStart) || ((*(UINT64 *)&ftStart == *(UINT64 *)&rhs.ftStart) && ((*(UINT64 *)&ftStop > *(UINT64 *)&rhs.ftStop) || ((*(UINT64 *)&ftStop == *(UINT64 *)&rhs.ftStop) && ((iChannelNum > rhs.iChannelNum) || ((iChannelNum == rhs.iChannelNum) && (strTitle > rhs.strTitle)))))));
		}
	};

	class Schedule : public vector<Show> {
	public:
		bool parse(istream &istr) {
			clear();
			istr.ignore(numeric_limits<streamsize>::max(), '\n');
			while (istr.good()) {
				string line;
				getline(istr, line);
				if (!line.empty()) {
					Show show;
					if (show.parse(line.c_str())) {
						push_back(show);
					}
					else {
						return false;
					}
				}
			}
			make_heap(begin(), end(), greater<Show>());
			return true;
		}

		template <typename C>
		void get_shows_at_time(const FILETIME *pft, const Options &opts, C &shows) const {
			FILETIME ftNow;
			if (pft == NULL) {
				GetSystemTimeAsFileTime(&ftNow);
				pft = &ftNow;
			}
			shows.clear();
			for (const_iterator it = begin(); it != end(); ++it) {
				if (it->is_selected(opts)) {
					if ((*(UINT64 *)pft >= *(UINT64 *)&it->ftStart) && (*(UINT64 *)pft < *(UINT64 *)&it->ftStop)) {
						shows.push_back(*it);
					}
				}
			}
		}

		template <typename C>
		void get_shows_at_next_time(const FILETIME *pft, const Options &opts, C &shows) const {
			FILETIME ftNow;
			if (pft == NULL) {
				GetSystemTimeAsFileTime(&ftNow);
				pft = &ftNow;
			}
			FILETIME ftSoonest = { ~0U, ~0U };
			shows.clear();
			for (const_iterator it = begin(); it != end(); ++it) {
				if (it->is_selected(opts)) {
					if (*(UINT64 *)pft < *(UINT64 *)&it->ftStart) {
						if (*(UINT64 *)&it->ftStart < *(UINT64 *)&ftSoonest) {
							shows.clear();
							shows.push_back(*it);
							ftSoonest = it->ftStart;
						}
						else if (*(UINT64 *)&it->ftStart == *(UINT64 *)&ftSoonest) {
							shows.push_back(*it);
						}
					}
				}
			}
		}
	};

	class WarningDialog {
	private:
		SageWake &app;
		HWND hwnd;
		FILETIME ftTimeout;
		int displayedSeconds;

		static INT_PTR WINAPI DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
			WarningDialog *pthis = (WarningDialog *)(LONG_PTR)GetWindowLongPtr(hwnd, DWLP_USER);
			switch (uMsg) {
				case WM_INITDIALOG: {
					pthis = (WarningDialog *)lParam;
					((LONG_PTR (WINAPI *)(HWND, int, LONG_PTR))SetWindowLongPtr)(hwnd, DWLP_USER, (LPARAM)pthis);
					FILETIME ftNow;
					GetSystemTimeAsFileTime(&ftNow);
					SendDlgItemMessage(hwnd, IDC_PROGRESS_SAND, PBM_SETRANGE32, 0, filetime_subtract_get_seconds(pthis->ftTimeout, ftNow));
					SetTimer(hwnd, 0, 100, NULL);
				}
				case WM_TIMER: {
					FILETIME ftNow;
					GetSystemTimeAsFileTime(&ftNow);
					int seconds = filetime_subtract_get_seconds(pthis->ftTimeout, ftNow);
					if (seconds <= 0) {
						pthis->app.log.event("Next suspend caused by SageWake.");
						KillTimer(hwnd, 0);
						SetSuspendState(pthis->app.opts.bUseHiber, FALSE, FALSE);
						SendMessage(hwnd, WM_CLOSE, 0, 0);
					}
					else if (seconds != pthis->displayedSeconds) {
						pthis->displayedSeconds = seconds;
						SendDlgItemMessage(hwnd, IDC_PROGRESS_SAND, PBM_SETPOS, seconds, 0);
						string str;
						format_seconds(seconds, str);
						SetDlgItemTextA(hwnd, IDC_STATIC_TIME, str.c_str());
					}
					return TRUE;
				}
				case WM_CLOSE: {
					delete pthis;
					return TRUE;
				}
				case WM_COMMAND: {
					switch (HIWORD(wParam)) {
						case BN_CLICKED: {
							switch (LOWORD(wParam)) {
								case IDCANCEL: {
									SendMessage(hwnd, WM_CLOSE, 0, 0);
									return TRUE;
								}
								default: {
									return FALSE;
								}
							}
						}
						default: {
							return FALSE;
						}
					}
				}
				default: {
					return FALSE;
				}
			}
		}

	public:
		WarningDialog(SageWake &app, FILETIME ftTimeout) : app(app), ftTimeout(ftTimeout), displayedSeconds(0) {
			hwnd = CreateDialogParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_WARN), NULL, DialogProc, (LPARAM)this);
		}

		~WarningDialog() {
			DestroyWindow(hwnd);
		}

		void show() {
			ShowWindow(hwnd, SW_SHOW);
		}
	};

	bool bPreventSleep;
	bool bDeniedSuspend;
	bool bSuspended;
	FILE *schedFileIO;
	ifstream schedFile;
	HANDLE hTimer;
	HICON hTrayIcon;
	HWND hwnd;
	FILETIME ftWakeTime;
	list<Show> nowRecording;
	list<Show> nextRecording;
	static UINT uTaskbarCreatedMsg;

	SageWake() : bPreventSleep(false), bDeniedSuspend(false), bSuspended(false), schedFileIO(NULL), hTimer(NULL), hwnd(NULL) {
		hTrayIcon = (HICON)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_APP), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
		*(UINT64 *)&ftWakeTime = 0;
	}

	~SageWake() {
		if (hTimer != NULL) {
			CloseHandle(hTimer);
		}
		if (hTrayIcon != NULL) {
			DestroyIcon(hTrayIcon);
		}
	}

	int run(int argc, const TCHAR **argv) {
		// Parse command line options
		if (!opts.initialize(argc, argv)) {
			return -1;
		}

		// Create the wake timer
		hTimer = CreateWaitableTimer(NULL, false, TEXT("SageWake"));
		if (GetLastError() == ERROR_ALREADY_EXISTS) {
			MessageBox(NULL, TEXT("An instance of SageWake is already running."), TEXT("SageWake"), MB_ICONEXCLAMATION);
			return -1;
		}

		// Get the full path to the schedule file
		TCHAR szPath[MAX_PATH];
		LPTSTR pszFilePart;
		if (SearchPath(NULL, opts.pszSchedFile, NULL, MAX_PATH, szPath, &pszFilePart) == 0) {
			MessageBox(NULL, TEXT("Unable to locate schedule file."), TEXT("SageWake"), MB_ICONEXCLAMATION);
			return -1;
		}

		// Open the schedule file
		schedFileIO = _tfsopen(szPath, TEXT("rtS"), _SH_DENYNO);
		if (schedFileIO == NULL) {
			MessageBox(NULL, TEXT("Unable to open schedule file."), TEXT("SageWake"), MB_ICONEXCLAMATION);
			return -1;
		}
		schedFile.rdbuf()->filebuf::basic_filebuf(schedFileIO);

		// Monitor directory containing schedule file for changes
		*(pszFilePart - 1) = 0;
		HANDLE hChange = FindFirstChangeNotification(szPath, false, FILE_NOTIFY_CHANGE_LAST_WRITE);
		if (hChange == INVALID_HANDLE_VALUE) {
			MessageBox(NULL, TEXT("Failed to monitor schedule file."), TEXT("SageWake"), MB_ICONEXCLAMATION);
			return -1;
		}

		// Create the schedule window
		hwnd = CreateDialogParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_SCHEDULE), NULL, _DialogProc, (LPARAM)this);
		add_tray_icon();

		// Open and set up the log file
		if (opts.pszLogFile != NULL) {
			log.open(opts.pszLogFile, opts.bAppendLog);
		}

		// Establish a reference schedule modification time
		struct _stat schedFileStat;
		_fstat(_fileno(schedFileIO), &schedFileStat);

		// Perform the initial parse of the schedule file
		schedFile.rdbuf()->pubseekpos(0);
		refresh_schedule(schedFile);

		HANDLE handles[] = { hTimer, hChange };
		for (;;) {
			switch (MsgWaitForMultipleObjects(sizeof(handles) / sizeof(*handles), handles, FALSE, INFINITE, QS_ALLINPUT)) {
				case WAIT_OBJECT_0: { // hTimer
					schedFile.rdbuf()->pubseekpos(0);
					schedFile.clear();
					refresh_schedule(schedFile);
					break;
				}
				case WAIT_OBJECT_0 + 1: { // hChange
					struct _stat updateStat;
					_fstat(_fileno(schedFileIO), &updateStat);
					if (updateStat.st_mtime != schedFileStat.st_mtime) {
						do {
							schedFileStat = updateStat;
							Sleep(1000);
							_fstat(_fileno(schedFileIO), &updateStat);
						} while (updateStat.st_mtime != schedFileStat.st_mtime);
						schedFile.rdbuf()->pubseekpos(0);
						schedFile.clear();
						refresh_schedule(schedFile);
					}
					FindNextChangeNotification(hChange);
					break;
				}
				case WAIT_OBJECT_0 + 2: { // message
					MSG msg;
					while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
						if (msg.message == WM_QUIT) {
							goto exit;
						}
						else if (!IsDialogMessage(hwnd, &msg)) {
							TranslateMessage(&msg);
							DispatchMessage(&msg);
						}
					}
					break;
				}
			}
		}
	exit:
		remove_tray_icon();
		DestroyWindow(hwnd);
		FindCloseChangeNotification(hChange);
		return 0;
	}

	static INT_PTR WINAPI _DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
		SageWake *pthis = (SageWake *)(LONG_PTR)GetWindowLongPtr(hwnd, DWLP_USER);
		switch (uMsg) {
			case WM_INITDIALOG: {
				pthis = (SageWake *)lParam;
				((LONG_PTR (WINAPI *)(HWND, int, LONG_PTR))SetWindowLongPtr)(hwnd, DWLP_USER, lParam);
				HICON hIcon = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_APP));
				SendMessage(hwnd, WM_SETICON, ICON_BIG, (LPARAM)hIcon);
				SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LPARAM)hIcon);
				SetWindowText(hwnd, TEXT("SageWake ") TEXT(APP_VERSION));
				int tabStops[] = { 10, 50 };
				SendDlgItemMessage(hwnd, IDC_LIST_SCHEDULE, LB_SETTABSTOPS, countof(tabStops), (LPARAM)tabStops);
				HWND hwndItem;
				hwndItem = GetDlgItem(hwnd, IDC_COMBO_SLEEPMODE);
				SendMessage(hwndItem, CB_ADDSTRING, 0, (LPARAM)TEXT("Every time this happens"));
				SendMessage(hwndItem, CB_ADDSTRING, 0, (LPARAM)TEXT("Just the next time this happens"));
				SendMessage(hwndItem, CB_ADDSTRING, 0, (LPARAM)TEXT("Only when I am not present"));
				SendMessage(hwndItem, CB_ADDSTRING, 0, (LPARAM)TEXT("Never"));
				EnableWindow(hwndItem, FALSE);
				hwndItem = GetDlgItem(hwnd, IDC_COMBO_SLEEPMETHOD);
				SendMessage(hwndItem, CB_ADDSTRING, 0, (LPARAM)TEXT("Suspend (S3)"));
				SendMessage(hwndItem, CB_ADDSTRING, 0, (LPARAM)TEXT("Hibernate (S4)"));
				SendMessage(hwndItem, CB_SETCURSEL, pthis->opts.bUseHiber ? 1 : 0, 0);
				return TRUE;
			}
			case WM_CLOSE: {
				ShowWindow(hwnd, SW_HIDE);
				return TRUE;
			}
			case WM_COMMAND: {
				switch (HIWORD(wParam)) {
					case 0: {
						switch (LOWORD(wParam)) {
							case ID_FILE_OPEN: {
								ShowWindow(hwnd, SW_SHOW);
								return TRUE;
							}
							case ID_APP_EXIT: {
								if (MessageBox(hwnd, TEXT("SageWake cannot wake your computer for scheduled recordings if it is not running. Are you sure you wish to close SageWake?"), TEXT("SageWake"), MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2) == IDYES) {
									PostQuitMessage(0);
								}
								return TRUE;
							}
							default: {
								return FALSE;
							}
						}
					}
					case CBN_SELCHANGE: {
						switch (LOWORD(wParam)) {
							case IDC_COMBO_SLEEPMETHOD: {
								pthis->opts.bUseHiber = (SendMessage((HWND)lParam, CB_GETCURSEL, 0, 0) == 1);
								return TRUE;
							}
							default: {
								return FALSE;
							}
						}
					}
					default: {
						return FALSE;
					}
				}
			}
			case WM_USER: {
				switch (lParam) {
					case WM_LBUTTONDOWN: {
						ShowWindow(hwnd, SW_SHOW);
						return TRUE;
					}
					case WM_CONTEXTMENU:
					case WM_RBUTTONUP: {
						POINT pt;
						GetCursorPos(&pt);
						HMENU hMenu = LoadMenu(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MENU_TRAY));
						SetForegroundWindow(hwnd);
						TrackPopupMenu(GetSubMenu(hMenu, 0), 0, pt.x, pt.y, 0, hwnd, NULL);
						DestroyMenu(hMenu);
						return TRUE;
					}
					default: {
						return FALSE;
					}
				}
			}
			case WM_POWERBROADCAST: {
				switch (wParam) {
					case PBT_APMQUERYSUSPEND: {
						if (pthis->bPreventSleep) {
							pthis->bDeniedSuspend = true;
							SetWindowLong(hwnd, DWLP_MSGRESULT, BROADCAST_QUERY_DENY);
							return TRUE;
						}
						return FALSE;
					}
					case PBT_APMQUERYSUSPENDFAILED: {
						if (pthis->bDeniedSuspend) {
							pthis->bDeniedSuspend = false;
						}
						else {
							pthis->do_execute();
						}
						return FALSE;
					}
					case PBT_APMSUSPEND: {
						if (!pthis->bSuspended) {
							pthis->bSuspended = true;
							pthis->do_terminate();
							pthis->log.event("<SystemSuspend />");
						}
						return FALSE;
					}
					case PBT_APMRESUMEAUTOMATIC:
					case PBT_APMRESUMECRITICAL:
					case PBT_APMRESUMESUSPEND: {
						if (pthis->bSuspended) {
							pthis->bSuspended = false;
							pthis->log.event("<SystemResume />");
							pthis->do_execute();
						}
						return FALSE;
					}
					default: {
						return FALSE;
					}
				}
			}
			default: {
				if (uMsg == uTaskbarCreatedMsg) {
					pthis->add_tray_icon();
					return TRUE;
				}
				return FALSE;
			}
		}
	}

	void add_tray_icon() {
		NOTIFYICONDATA nid;
		nid.cbSize = sizeof(NOTIFYICONDATA);
		nid.hWnd = hwnd;
		nid.uID = 0;
		nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
		nid.uCallbackMessage = WM_USER;
		nid.hIcon = hTrayIcon;
		_tcsncpy_s(nid.szTip, TEXT("SageWake"), _TRUNCATE);
		Shell_NotifyIcon(NIM_ADD, &nid);
	}

	void remove_tray_icon() {
		NOTIFYICONDATAA nid;
		nid.cbSize = sizeof(NOTIFYICONDATAA);
		nid.hWnd = hwnd;
		nid.uID = 0;
		nid.uFlags = 0;
		Shell_NotifyIconA(NIM_DELETE, &nid);
	}

	void balloon_tray_icon(LPCSTR pszMsg) {
		NOTIFYICONDATAA nid;
		nid.cbSize = sizeof(NOTIFYICONDATAA);
		nid.hWnd = hwnd;
		nid.uID = 0;
		nid.uFlags = NIF_INFO;
		strncpy_s(nid.szInfo, pszMsg, _TRUNCATE);
		nid.uTimeout = 0;
		strncpy_s(nid.szInfoTitle, "SageWake", _TRUNCATE);
		nid.dwInfoFlags = NIIF_INFO;
		Shell_NotifyIconA(NIM_MODIFY, &nid);
	}

	void refresh_schedule(istream &istr) {
		Schedule sched;
		sched.parse(istr);
		FILETIME ftNow;
		GetSystemTimeAsFileTime(&ftNow);
		list<Show> nowRecording, nextRecording;
		sched.get_shows_at_time(&ftNow, opts, nowRecording);
		sched.get_shows_at_next_time(&ftNow, opts, nextRecording);
		string strEvent, str;
		HWND hwndList = GetDlgItem(hwnd, IDC_LIST_SCHEDULE);
		SendMessage(hwndList, LB_RESETCONTENT, 0, 0);
		for (Schedule::const_iterator it = sched.begin(); it != sched.end(); ++it) {
			it->to_gui_string(str, opts, true);
			SendMessageA(hwndList, LB_ADDSTRING, 0, (LPARAM)str.c_str());
		}
		if ((nowRecording.size() != this->nowRecording.size()) || !equal(nowRecording.begin(), nowRecording.end(), this->nowRecording.begin())) {
			this->nowRecording.assign(nowRecording.begin(), nowRecording.end());
			if (log.is_open()) {
				strEvent = "\r\n\t<NowRecording>";
				for (list<Show>::const_iterator it = nowRecording.begin(); it != nowRecording.end(); ++it) {
					strEvent += "\r\n\t\t";
					it->to_xml(str);
					strEvent += str;
				}
				strEvent += "\r\n\t</NowRecording>";
			}
			if (!nowRecording.empty()) {
				str = "SageTV now recording:";
				for (list<Show>::const_iterator it = nowRecording.begin(); it != nowRecording.end(); ++it) {
					str.push_back('\n');
					str.append(it->strTitle);
				}
				balloon_tray_icon(str.c_str());
			}
		}
		if ((nextRecording.size() != this->nextRecording.size()) || !equal(nextRecording.begin(), nextRecording.end(), this->nextRecording.begin())) {
			this->nextRecording.assign(nextRecording.begin(), nextRecording.end());
			if (log.is_open()) {
				strEvent += "\r\n\t<NextRecording>";
				for (list<Show>::const_iterator it = nextRecording.begin(); it != nextRecording.end(); ++it) {
					strEvent += "\r\n\t\t";
					it->to_xml(str);
					strEvent += str;
				}
				strEvent += "\r\n\t</NextRecording>";
			}
			if (nowRecording.empty() && !nextRecording.empty()) {
				str = "SageTV next recording:";
				for (list<Show>::const_iterator it = nextRecording.begin(); it != nextRecording.end(); ++it) {
					str.push_back('\n');
					str.append(it->strTitle);
				}
				balloon_tray_icon(str.c_str());
			}
		}
		bool bPreventSleep = !nowRecording.empty();
		if (bPreventSleep) {
			FILETIME ftSoonest = { ~0U, ~0U };
			for (list<Show>::const_iterator it = nowRecording.begin(); it != nowRecording.end(); ++it) {
				if (*(UINT64 *)&it->ftStop < *(UINT64 *)&ftSoonest) {
					ftSoonest = it->ftStop;
				}
			}
			SetWaitableTimer(hTimer, (LARGE_INTEGER *)&ftSoonest, 0, NULL, NULL, FALSE);
		}
		else if (!nextRecording.empty()) {
			FILETIME ftWakeTime = filetime_add_seconds(nextRecording.front().ftStart, -opts.iLeadTime);
			if (*(UINT64 *)&ftWakeTime > *(UINT64 *)&ftNow) {
				SetWaitableTimer(hTimer, (LARGE_INTEGER *)&ftWakeTime, 0, NULL, NULL, TRUE);
				if (log.is_open() && (*(UINT64 *)&ftWakeTime != *(UINT64 *)&this->ftWakeTime)) {
					this->ftWakeTime = ftWakeTime;
					FILETIME ftLocalWakeTime;
					FileTimeToLocalFileTime(&ftWakeTime, &ftLocalWakeTime);
					SYSTEMTIME stLocalWakeTime;
					FileTimeToSystemTime(&ftLocalWakeTime, &stLocalWakeTime);
					strEvent += "\r\n\t<SetWakeUp time=\"";
					get_short_date(&stLocalWakeTime, str);
					strEvent += str;
					strEvent.push_back(' ');
					get_short_time(&stLocalWakeTime, str);
					strEvent += str;
					strEvent += "\" />";
				}
			}
			else {
				SetWaitableTimer(hTimer, (LARGE_INTEGER *)&nextRecording.front().ftStart, 0, NULL, NULL, FALSE);
				bPreventSleep = true;
			}
		}
		if (bPreventSleep != this->bPreventSleep) {
			this->bPreventSleep = bPreventSleep;
			if (bPreventSleep) {
				SetThreadExecutionState(ES_SYSTEM_REQUIRED | ES_CONTINUOUS);
				if (log.is_open()) {
					strEvent += "\r\n\t<SetAllowSleep allow=\"false\" />";
				}
			}
			else {
				SetThreadExecutionState(ES_CONTINUOUS);
				if (log.is_open()) {
					strEvent += "\r\n\t<SetAllowSleep allow=\"true\" />";
				}
			}
		}
		if (nowRecording.empty() && !bPreventSleep && IsSystemResumeAutomatic() && !opts.bNoResuspend) {
			new WarningDialog(*this, filetime_add_seconds(ftNow, opts.iWarnTime));
		}
		if (log.is_open() && (strEvent.length() != 0)) {
			strEvent += "\r\n";
			log.event(strEvent.c_str());
		}
	}

	void do_terminate() {
		EnumWindows(_terminate_enum_proc, (LPARAM)this);
	}

	static BOOL WINAPI _terminate_enum_proc(HWND hwnd, LPARAM lParam) {
		return ((SageWake *)lParam)->terminate_enum_proc(hwnd);
	}

	BOOL terminate_enum_proc(HWND hwnd) {
		DWORD dwThreadId, dwProcessId;
		HANDLE hProcess;
		TCHAR sz[64];

		GetClassName(hwnd, sz, 64);
		if (find(opts.lstTerminate.begin(), opts.lstTerminate.end(), sz) == opts.lstTerminate.end()) {
			return TRUE;
		}
#ifdef UNICODE
		char szA[64];
		wcstombs_s(NULL, szA, sz, 64);
#else
#define szA sz
#endif
		dwThreadId = GetWindowThreadProcessId(hwnd, &dwProcessId);
		hProcess = OpenProcess(SYNCHRONIZE, FALSE, dwProcessId);
		if (hProcess != NULL) {
			PostThreadMessage(dwThreadId, WM_QUIT, 0, 0);
			if (WaitForSingleObject(hProcess, 5000) == WAIT_OBJECT_0) {
				CloseHandle(hProcess);
				stringstream stm;
				stm << "<QuitProcessCleanly class=\"" << szA << "\" id=\"" << dwProcessId << "\" />";
				log.event(stm.str().c_str());
				return TRUE;
			}
			CloseHandle(hProcess);
		}
		hProcess = OpenProcess(PROCESS_TERMINATE | SYNCHRONIZE, FALSE, dwProcessId);
		if (hProcess != NULL) {
			if (TerminateProcess(hProcess, 0)) {
				WaitForSingleObject(hProcess, INFINITE);
				CloseHandle(hProcess);
				stringstream stm;
				stm << "<TerminatedProcessForcefully class=\"" << sz << "\" id=\"" << dwProcessId << "\" />";
				log.event(stm.str().c_str());
				return TRUE;
			}
			CloseHandle(hProcess);
		}
#undef szA
		return TRUE;
	}

	void do_execute() {
		PROCESS_INFORMATION pi;
		STARTUPINFO si = { sizeof(STARTUPINFO), 0 };

		for (list<_tstring>::const_iterator it = opts.lstExecute.begin(); it != opts.lstExecute.end(); ++it) {
			if (CreateProcess(it->c_str(), NULL, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
				CloseHandle(pi.hProcess);
				CloseHandle(pi.hThread);
#ifdef UNICODE
				string str;
				insert_iterator<string> ins(str, str.end());
				transform(it->begin(), it->end(), ins, wctob);
#else
#define str (*it)
#endif
				stringstream stm;
				stm << "<CreatedProcess path=\"" << str << "\" id=\"" << pi.dwProcessId << "\" />";
				log.event(stm.str().c_str());
#undef str
			}
		}
	}
};

UINT SageWake::uTaskbarCreatedMsg = RegisterWindowMessage(TEXT("TaskbarCreated"));

int _tmain(int argc, TCHAR **argv) {
	{
		INITCOMMONCONTROLSEX icce = { sizeof(INITCOMMONCONTROLSEX), ICC_PROGRESS_CLASS };
		InitCommonControlsEx(&icce);
	}
	SageWake app;
	return app.run(argc, const_cast<const TCHAR **>(argv));
}
